/*
 * This file is (c) Copyright Zarafa, The Netherlands, 2007
 *
 * Use of this file and the source contained within it is subject to the
 * license of the package it belongs to. See the LICENSE file in the
 * top-level directory of the package.
 *
 */

#include "platform.h"
#include "stringutil.h"

// Damn windows header defines max which break C++ header files
#undef max

#include <string>
#include <fstream>
#include <iostream>
#include <stdlib.h>

// vmime
#include "vmime/vmime.hpp"
#ifdef _WIN32
#include "vmime/platforms/windows/windowsHandler.hpp"
#else
#include "vmime/platforms/posix/posixHandler.hpp"
#endif

// mapi
#include <mapix.h>
#include <mapiutil.h>
#include <mapiext.h>

// inetmapi
#include "inetmapi.h"
#include "VMIMEToMAPI.h"
#include "MAPIToVMIME.h"
#include "ECVMIMEUtils.h"
#include "ECMapiUtils.h"


#define STREAM_BUFF_SIZE	4096		// size of buffer used for looping body and attachments into open property stream
#define LINE_FEED			0x0a		// '\n'
#define CARRIAGE_RETURN		0x0d		// '\r'


ECSender::ECSender(ECLogger *newlpLogger, std::string strSMTPHost) {
	lpLogger = newlpLogger;
	if (!lpLogger) {
		lpLogger = new ECLogger_Null();
		delete_logger = true;
	} else {
		delete_logger = false;
	}
	smtpresult = 0;
	smtphost = strSMTPHost;
}

ECSender::~ECSender() {
	if (delete_logger)
		delete lpLogger;
}

int ECSender::getSMTPResult() {
	return smtpresult;
}

const char* ECSender::getErrorString() {
	return error.c_str();
}

void ECSender::setError(std::string newError) {
	error = newError;
}

INETMAPI_API ECSender* CreateSender(ECLogger *lpLogger, std::string smtp) {
	return new ECVMIMESender(lpLogger, smtp);
}

// Read stream lpStream and set properties on open lpMessage object 
INETMAPI_API HRESULT IMToMAPI(IMAPISession *lpSession, IMessage *lpMessage, string input, delivery_options dopt, ECLogger *lpLogger)
{
	HRESULT	hResult = hrSuccess;
	LPADRBOOK lpAdrBook = NULL;
	VMIMEToMAPI *VMToM = NULL;

	lpSession->OpenAddressBook(0, NULL, AB_NO_DIALOG, &lpAdrBook);
	// don't care if addressbook fails, we have fallbacks

	VMToM = new VMIMEToMAPI(lpAdrBook, lpLogger, dopt);

	// prepare vmime for working on my platform
#ifdef _WIN32
    vmime::platformDependant::setHandler<vmime::platforms::windows::windowsHandler>();
#else
    vmime::platformDependant::setHandler<vmime::platforms::posix::posixHandler>();
#endif

	// fill mapi object from buffer
	hResult = VMToM->convertVMIMEToMAPI(input, lpMessage);
	delete VMToM;

	if (lpAdrBook)
		lpAdrBook->Release();

	return hResult;
}

// Read properties from lpMessage object and fill a buffer with internet rfc822 format message
INETMAPI_API HRESULT IMToINet(IMAPISession *lpSession, IMessage *lpMessage, char** lppbuf, sending_options sopt, ECLogger *lpLogger)
{
    HRESULT hr = hrSuccess;
    std::ostringstream oss;
    char *lpszData = NULL;
    
    hr = IMToINet(lpSession, lpMessage, oss, sopt, lpLogger);
    
    if(hr != hrSuccess)
        goto exit;
        
    lpszData = new char[oss.str().size()+1];
    strcpy(lpszData, oss.str().c_str());
    
    *lppbuf = lpszData;
exit:    
    return hr;
}

INETMAPI_API HRESULT IMToINet(IMAPISession *lpSession, IMessage *lpMessage, std::ostream &os, sending_options sopt, ECLogger *lpLogger)
{
	HRESULT			hr			= hrSuccess;
	LPSPropValue	lpTime		= NULL;
	LPSPropValue	lpMessageId	= NULL;
	MAPIToVMIME*	mToVM		= new MAPIToVMIME(lpSession, lpLogger, sopt);
	vmime::message*	lpVMMessage	= NULL;

	vmime::utility::outputStreamAdapter adapter(os);
	
#ifdef _WIN32
	vmime::platformDependant::setHandler<vmime::platforms::windows::windowsHandler>();
#else
	vmime::platformDependant::setHandler<vmime::platforms::posix::posixHandler>();
#endif

	hr = mToVM->convertMAPIToVMIME(lpMessage, &lpVMMessage);
	if (hr != hrSuccess)
		goto exit;

	try {
		// vmime messageBuilder has set Date header to now(), so we overwrite it.
		if (HrGetOneProp(lpMessage, PR_CLIENT_SUBMIT_TIME, &lpTime) == hrSuccess) {
			lpVMMessage->getHeader()->Date().setValue(FiletimeTovmimeDatetime(lpTime->Value.ft));
		}
		// else, try PR_MESSAGE_DELIVERY_TIME, maybe other timestamps?

		if (HrGetOneProp(lpMessage, PR_INTERNET_MESSAGE_ID, &lpMessageId) == hrSuccess) {
			lpVMMessage->getHeader()->MessageId().setValue(lpMessageId->Value.lpszA);
		}

		// generate message and put into buffer.
		lpVMMessage->generate(adapter);
	}
	catch (vmime::exception& e) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	catch (std::exception& e) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	
exit:
	if (lpTime)
		MAPIFreeBuffer(lpTime);

	if (lpMessageId)
		MAPIFreeBuffer(lpMessageId);

	if (lpVMMessage)
		delete lpVMMessage;
		
	delete mToVM;

	return hr;
}


// Read properties from lpMessage object and write to lpStream with internet rfc2822 format message
// then send it using smtp://lpSMTP
INETMAPI_API HRESULT IMToINet(IMAPISession *lpSession, IMessage *lpMessage, ECSender *mailer_base, sending_options sopt, ECLogger *lpLogger)
{
	HRESULT			hr				= hrSuccess;
	MAPIToVMIME		*mToVM			= new MAPIToVMIME(lpSession, lpLogger, sopt);
	vmime::message	*lpVMMessage 	= NULL;
	ECVMIMESender	*mailer = (ECVMIMESender*)mailer_base;

#ifdef _WIN32
	vmime::platformDependant::setHandler<vmime::platforms::windows::windowsHandler>();
#else
	vmime::platformDependant::setHandler<vmime::platforms::posix::posixHandler>();
#endif
	
	hr = mToVM->convertMAPIToVMIME(lpMessage, &lpVMMessage);

	if (hr != hrSuccess) {
		mailer->setError("Conversion error: "+stringify(hr,true));
		goto exit;
	}

	try {
		// vmime messageBuilder has set Date header to now()
		lpVMMessage->getHeader()->MessageId().setValue(vmime::messageId::generateId());
	}
	catch (vmime::exception& e) {
	    mailer->setError(e.what());
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	catch (std::exception& e) {
	    mailer->setError(e.what());
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}
	catch (...) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

// 	hr = mailer->printMail(lpVMMessage);
	hr = mailer->sendMail(lpVMMessage);

exit:
	if (lpVMMessage)
		delete lpVMMessage;		// allocated by vmime::messageBuilder
	delete mToVM;

	return hr;
}
